#import "XeeQuicktimeLoader.h"

#import "XeeBitmapImage.h"
#import "XeeNSImageLoader.h"



void set_qt_depth(XeeImage *image,int qtdepth)
{
	if(qtdepth<=8) [image setDepthIndexed:(1<<qtdepth)];
	else if(qtdepth>32) [image setDepthGrey:(1<<(qtdepth-32))];
	else if(qtdepth==15) [image setDepthRGB:15];
	else if(qtdepth==16) [image setDepthRGBA:16];
	else if(qtdepth==24) [image setDepthRGB:24];
	else if(qtdepth==32) [image setDepthRGBA:32];
}


@implementation XeeQuicktimeImage

-(id)initWithFile:(NSString *)name firstBlock:(NSData *)block attributes:(NSDictionary *)attributes;
{
	if(self=[super initWithFile:name firstBlock:block attributes:attributes])
	{
		gi=NULL;

		if(
			[[filename pathExtension] caseInsensitiveCompare:@"bmp"]!=NSOrderedSame
			&&[attrs fileHFSTypeCode]!='BMP '&&[attrs fileHFSTypeCode]!='BMP '
			&&[[filename pathExtension] caseInsensitiveCompare:@"tga"]!=NSOrderedSame
			&&[attrs fileHFSTypeCode]!='TPIC'
			&&[[filename pathExtension] caseInsensitiveCompare:@"sgi"]!=NSOrderedSame
			&&[[filename pathExtension] caseInsensitiveCompare:@"rgb"]!=NSOrderedSame
			&&[attrs fileHFSTypeCode]!='.SGI'
		) // explicitly exclude BMP, TGA and SGI - what the hell, Quicktime?
		{
			NSURL *url=[NSURL fileURLWithPath:filename];
			FSRef fsref;
			if(CFURLGetFSRef((CFURLRef)url,&fsref))
			{
				FSSpec fsspec;
				if(FSGetCatalogInfo(&fsref,kFSCatInfoNone,NULL,NULL,&fsspec,NULL)==noErr)
				{
					if(GetGraphicsImporterForFile(&fsspec,&gi)==noErr)
					{
						ImageDescriptionHandle desc;

						if(GraphicsImportGetImageDescription(gi,&desc)==noErr)
						{
							width=(*desc)->width;
							height=(*desc)->height;

							switch((*desc)->cType)
							{
								case kJPEGCodecType: [self setFormat:@"JPEG"]; break;
								default:
								{
									NSString *loc=[[[NSBundle mainBundle] preferredLocalizations] objectAtIndex:0];

									if([loc isEqual:@"Japanese"]) [self setFormat: [[[NSString alloc] initWithBytes:(*desc)->name+1 length:(*desc)->name[0] encoding:CFStringConvertEncodingToNSStringEncoding(kCFStringEncodingMacJapanese)] autorelease]];
									else [self setFormat:[[[NSString alloc] initWithBytes:(*desc)->name+1 length:(*desc)->name[0] encoding:NSMacOSRomanStringEncoding] autorelease]];
								}
								break;
							}

							int qtdepth=(*desc)->depth;
							set_qt_depth(self,qtdepth);

							if(qtdepth==32) [self setTransparent:YES];

							DisposeHandle((Handle)desc);

							nextindex=0;

							[self setState:XeeImageStateLoading];

							return self;
						}
					}
				}
			}
		}

		[self release];
	}

	return nil;
}

-(void)dealloc
{
	if(gi) CloseComponent(gi);

	[super dealloc];
}


static OSErr progress_func(short message,Fixed completeness,long refcon)
{
	XeeQuicktimeImage *image=(XeeQuicktimeImage *)refcon; // ow ow ow, 64-bit issues!

	if([image hasBeenStopped]) return codecAbortErr;

/*	static int i=0;
	if((i++)%32==0)
	{
		XeeBitmapImage *subimage=[image finalSubImage];
		[subimage triggerUpdateAction];
	}*/

	return noErr;
}

-(void)load
{
	unsigned long count;
	GraphicsImportGetImageCount(gi,&count);

	[self loadFromIndex:1 count:count];
}

-(void)loadThumbnail
{
	if(GraphicsImportSetImageIndexToThumbnail(gi)==noErr)
	{
		unsigned long index;
		GraphicsImportGetImageIndex(gi,&index);

		[self loadFromIndex:index count:1];
	}
	else
	{
		[self load];
	}
}

-(void)loadFromIndex:(int)startindex count:(int)count
{
	stop=NO;

	EnterMoviesOnThread(0);

	CGrafPtr origport;
	GDHandle origdev;
	GetGWorld(&origport,&origdev);

	struct ICMProgressProcRecord progrec={progress_func,(long)self};
	GraphicsImportSetProgressProc(gi,&progrec);

	GraphicsImportSetQuality(gi,codecLosslessQuality);

	while(nextindex<count)
	{
		if(stop) break;

		if(GraphicsImportSetImageIndex(gi,nextindex+startindex)==noErr)
		{
			ImageDescriptionHandle desc;
			if(GraphicsImportGetImageDescription(gi,&desc)==noErr)
			{
				int framewidth=(*desc)->width;
				int frameheight=(*desc)->height;
				int framedepth=(*desc)->depth;
				DisposeHandle((Handle)desc);

				int type;

				if(framedepth>32) type=XeeBitmapTypeLuma8;
				else if([format isEqual:@"JPEG"]&&[[NSUserDefaults standardUserDefaults] integerForKey:@"jpegYUV"]==1) type=XeeBitmapTypeYUV422;
				else if(framedepth==32||framedepth==16) type=XeeBitmapTypeARGB32;
				else type=XeeBitmapTypeRGB24;

				XeeBitmapImage *image=[[[XeeBitmapImage alloc] initWithType:type width:framewidth height:frameheight] autorelease];

				if(image)
				{
					if(framedepth==32) [image setTransparent:YES];

					set_qt_depth(image,framedepth);

					if(GraphicsImportSetGWorld(gi,[image gworld],NULL)==noErr)
					{
//						[self addSubImage:image];

						ComponentResult res=GraphicsImportDraw(gi);
						if(res==noErr)
						{
							if(type==XeeBitmapTypeYUV422) [image fixYUVGamma];
							if(type==XeeBitmapTypeLuma8) [image invertChannels];

							[self addSubImage:image];
							[image triggerUpdateAction];
						}
						else
						{
							if(!stop) [subimages removeAllObjects];
							break;
						}
					}
				}
			}
		}
		nextindex++;
	}

	SetGWorld(origport,origdev);

	ExitMoviesOnThread();

	if(stop) return;

	CloseComponent(gi);
	gi=NULL;

	if(![subimages count]) // failed to load any frames, ask XeeNSImage for help
	{
		NSImage *image=[[[NSImage alloc] initWithContentsOfFile:filename] autorelease];
		NSArray *images=[XeeNSImage convertRepresentations:[image representations]];
		if(images) [self addSubImages:images];

		if(![subimages count])
		{
			[self setState:XeeImageStateFailed];
			return;
		}
	}

	[self setState:XeeImageStateLoaded];
	[self triggerUpdateAction];
}

-(XeeBitmapImage *)finalSubImage { return [subimages objectAtIndex:[subimages count]-1]; }



+(void)load
{
	EnterMovies();
}

+(NSArray *)fileTypes
{
	return [NSArray arrayWithObjects:
		@"jpg", // JPEG Image
		@"jpeg",
		@"jpe",
		@"'JPEG'",

		@"png", // Portable Network Graphics
		@"'PNG '", 
		@"'PNGf'", 

		@"gif", // Graphics Interchange Format
		@"'GIFf'",
		@"'GIF '",

		@"tif", // TIFF Image
		@"tiff", 
		@"'TIFF'", 

		@"bmp", // Windows Bitmap Image
		@"dib", //?
		@"'BMP '", 
		@"'BMPf'", 
		@"'BMPp'", 

		@"psd", // Adobe Photoshop Image
		@"'8BPS'",

		@"tga", // Targa Image
		@"'TPIC'", 

		@"jp2", // JPEG 2000 Image
		@"'jp2 '", 

		@"pict", // PICT Image
		@"pct", 
		@"pic",
		@"'PICT'", 

		@"fpx", // FlashPix Image
		@"'FPix'", 

		@"qtif", // Apple Quicktime Image
		@"qti",
		@"qif", 
		@"'qtif'", 

		@"sgi", // Silicon Graphics Image
		@"'SGI '", 
		@"'.SGI'", 

		@"pntg", // Apple MacPaint image
		@"'PNTG'", 

		@"mac", //?

		nil
	];
}

@end
